'''
    db-patch applies static and dynamic patches to databases using ODBC or
    native connectors.
    Copyright (C) 2012, 2013  Miguel A. Martinez Pinedo

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301  USA
'''

import os
import shutil
import re
import time
import datetime
import ConfigParser

import sqlite3

from dbpatch.color import *


class HistoryCommand(object):

    EXPORT   = 'export'
    DELETE   = 'delete'
    LIST     = 'list'
    SEE      = 'see'

    def __init__(self, command, arguments, subcommand=None):
        self._command    = command
        self._arguments  = arguments
        self._subcommand = subcommand

    def __repr__(self):
        if self._subcommand:
            return 'Cmd:[%s] [%s]. Args: %s' % (self._command,
                                                self._subcommand,
                                                self._arguments)
        return 'Cmd:[%s] Args: %s' % (self._command, self._arguments)

    def __get_command(self):
        return self._command

    def __get_subcommand(self):
        return self._subcommand

    def __get_arguments(self):
        return self._arguments

    def is_see_command(self):
        return (self._command == HistoryCommand.SEE)

    def is_export_command(self):
        return (self._command == HistoryCommand.EXPORT)

    def is_list_command(self):
        return (self._command == HistoryCommand.LIST)

    def is_delete_command(self):
        return (self._command == HistoryCommand.DELETE)

    command    = property(fget=__get_command)
    subcommand = property(fget=__get_subcommand)
    arguments  = property(fget=__get_arguments)


class CommandParser(object):

    DATE = '\d\d\d\d-\d\d-\d\d'

    def __init__(self, key, the_re, keywords=[]):
        self._key = key
        self._re  = the_re
        self._keywords = keywords

    def __get_key(self):
        return self._key

    def __get_re(self):
        return self._re

    def __get_keywords(self):
        return self._keywords

    key = property(fget=__get_key)
    re  = property(fget=__get_re)
    keywords = property(fget=__get_keywords)


class HistoryCommandParser(object):

    class UnrecognizedCommand(Exception):
        def __init__(self, message):
            Exception.__init__(self, message)

    class MalformedCommand(Exception):
        def __init__(self, message):
            Exception.__init__(self, message)

    class GenericParser(object):

        def get_key(self):
            raise NotImplementedError

        def get_commands(self):
            raise NotImplementedError

        def parse(self, raw_command):
            for command in self.get_commands():
                if command.key not in raw_command:
                    continue

                pattern = re.compile(command.re)
                if not pattern.match(raw_command):
                    raise HistoryCommandParser.MalformedCommand(raw_command)

                return HistoryCommandParser.create_history_command(
                    raw_command, command.keywords)

            raise HistoryCommandParser.UnrecognizedCommand(raw_command)

    class ListCommandParser(GenericParser):
        LIST_KEY = 'list '
        COMMANDS = [
            CommandParser(' all', '(\s)*list(\s)+all(\s)*', ['list', 'all']),
            CommandParser(' db ', '(\s)*list(\s)+db(\s)+(\S)+(\s)*',
                          ['list', 'db']),
            CommandParser(' host ', '(\s)*list(\s)+host(\s)+(\S)+(\s)*',
                          ['list', 'host']),
            CommandParser(
                ' since ',
                '(\s)*list(\s)+since(\s)+%s(\s)*' % CommandParser.DATE,
                ['list', 'since']),
            CommandParser(
                ' until ',
                '(\s)*list(\s)+until(\s)+%s(\s)*' % CommandParser.DATE,
                ['list', 'until']),
            CommandParser(
                'between',
                '(\s)*list(\s)+between(\s)+(\d)+(\s)+(\d)+(\s)*',
                ['list', 'between'])]

        def get_key(self):
            return HistoryCommandParser.ListCommandParser.LIST_KEY

        def get_commands(self):
            return HistoryCommandParser.ListCommandParser.COMMANDS

    class DeleteCommandParser(GenericParser):
        DELETE_KEY = 'delete '
        COMMANDS   = [
            CommandParser(' entries', '(\s)*delete(\s)+entries((\s)+(\d)+(\s)*)+',
                          ['delete', 'entries']),
            CommandParser(' all', '(\s)*delete(\s)+all(\s)*',
                          ['delete', 'all']),
            CommandParser(' db ', '(\s)*delete(\s)+db(\s)+(\S)+(\s)*',
                          ['delete', 'db']),
            CommandParser(' host ', '(\s)*delete(\s)+host(\s)+(\S)+(\s)*',
                          ['delete', 'host']),
            CommandParser(
                ' since ',
                '(\s)*delete(\s)+since(\s)+%s(\s)*' % CommandParser.DATE,
                ['delete', 'since']),
            CommandParser(
                ' until ',
                '(\s)*delete(\s)+until(\s)+%s(\s)*' % CommandParser.DATE,
                ['delete', 'until']),
            CommandParser(
                'between',
                '(\s)*delete(\s)+between(\s)+(\d)+(\s)+(\d)+(\s)*',
                ['delete', 'between'])]

        def get_key(self):
            return HistoryCommandParser.DeleteCommandParser.DELETE_KEY

        def get_commands(self):
            return HistoryCommandParser.DeleteCommandParser.COMMANDS

    class OtherCommandParser(GenericParser):

        COMMANDS = [
            CommandParser('see ', '(\s)*see(\s)+(\d)+(\s)*', ['see']),
            CommandParser('export ',
                          '(\s)*export(\s)+[(\d)+(\s)+]+(\s)+(\S)+(\s)*',
                          ['export'])]

        def get_commands(self):
            return HistoryCommandParser.OtherCommandParser.COMMANDS

    @staticmethod
    def parse(raw_command):
        if HistoryCommandParser.ListCommandParser.LIST_KEY in raw_command:
            cmd_parser = HistoryCommandParser.ListCommandParser()
        elif HistoryCommandParser.DeleteCommandParser.DELETE_KEY in raw_command:
            cmd_parser = HistoryCommandParser.DeleteCommandParser()
        else:
            cmd_parser = HistoryCommandParser.OtherCommandParser()

        return cmd_parser.parse(raw_command)


    @staticmethod
    def create_history_command(raw_command, keywords):
        args = filter(lambda arg : arg not in keywords, raw_command.split())
        cmd  = keywords[0]
        subcmd = keywords[1] if len(keywords) > 1 else None
        return HistoryCommand(cmd, args, subcmd)


class HistoryFormatter(object):

    @staticmethod
    def line(sizes=[5,10,10,20,25,19]):
        return '+%s+%s+%s+%s+%s+%s+' % ('-' * (sizes[0]),
                                        '-' * (sizes[1]),
                                        '-' * (sizes[2]),
                                        '-' * (sizes[3]),
                                        '-' * (sizes[4]),
                                        '-' * (sizes[5]))

    @staticmethod
    def header(sizes=[5,10,10,20,25,19]):
        header = '|%-*s|%-*s|%-*s|%-*s|%-*s|%-*s|' % (
            sizes[0], 'Id',
            sizes[1], 'Hostname',
            sizes[2], 'Db Name',
            sizes[3], 'Patch File',
            sizes[4], 'SQL File',
            sizes[5], 'Date')

        line = HistoryFormatter.line(sizes)

        return '%s\n%s\n%s' % (line, header, line)


    @staticmethod
    def row(the_id, host, db_name, patch_file, out_file,
            date, sizes=[5,10,10,20,25,19]):
        return '|%-*s|%-*s|%-*s|%-*s|%-*s|%-*s|' % (
            sizes[0], str(the_id),
            sizes[1], host,
            sizes[2], db_name,
            sizes[3], patch_file,
            sizes[4], out_file,
            sizes[5], date)


class HistoricEntry(object):

    def __init__(self, db_name, patch_file, host='localhost', out_file=None):
        self._host    = host
        self._db_name = db_name
        self._patch_file = patch_file
        self._out_file   = out_file
        self._date = HistoricEntry.create_datetime_string()
        self._id   = None

    @staticmethod
    def create_datetime_string():
        year, month, day, hour, mins, \
            secs, wday, yday, isdst = time.localtime()
        # 'YYYY-MM-DD HH:MM:SS'
        return '%04d-%02d-%02d %02d:%02d:%02d' % (year, month, day,
                                                  hour, mins, secs)

    def __repr__(self):
        return '%-05s %-10s %10s %20s %20s %20s' % (str(self.__get_id()),
                                                    self._host,
                                                    self._db_name,
                                                    self._patch_file,
                                                    self.__get_out_file(),
                                                    self._date)
    def to_string(self, sizes=[5,10,10,20,25,19]):
        return HistoryFormatter.row(self.__get_id(),
                                    self._host,
                                    self._db_name,
                                    self._patch_file,
                                    self.__get_out_file(),
                                    self._date,
                                    sizes)


    def __get_host(self):
        return self._host

    def __get_db_name(self):
        return self._db_name

    def __get_patch_file(self):
        return self._patch_file

    def __get_out_file(self):
        if not self._out_file:
            return ''
        return self._out_file

    def __get_date(self):
        return self._date

    def __set_date(self, date):
        self._date = date

    def __get_id(self):
        if not self._id:
            return ''
        return self._id

    def __set_id(self, the_id):
        self._id = the_id

    host       = property(fget=__get_host)
    db_name    = property(fget=__get_db_name)
    patch_file = property(fget=__get_patch_file)
    out_file   = property(fget=__get_out_file)
    date       = property(fget=__get_date, fset=__set_date)
    the_id     = property(fget=__get_id, fset=__set_id)


class HistoricalConfigFile(object):

    SECTION    = 'historical_db'

    BASE_PATH  = 'base_path'
    DB_NAME    = 'db_name'
    PATCH_PATH = 'patch_path'
    SQL_PATH   = 'sql_path'

    class RequiredSection(Exception):
        def __init__(self, message):
            Exception.__init__(self, message)

    class RequiredOption(Exception):
        def __init__(self, message):
            Exception.__init__(self, message)

    class DefaultValues(object):
        def __init__(self):
            self.base_path  = os.path.realpath(
                os.path.expanduser('~/.db-patch/db'))
            self.db_name    = 'history.sqlite'
            self.patch_path = 'patches'
            self.sql_path   = 'sqls'

    def __init__(self, the_file):
        self._parser  = None
        self._default = None

        full_name = os.path.realpath(os.path.expanduser(the_file))
        if not os.path.exists(full_name) or not os.path.isfile(full_name):
            self._default = HistoricalConfigFile.DefaultValues()
            return

        self._parser = ConfigParser.ConfigParser()
        self._parser.read(full_name)

        if not self._parser.has_section(HistoricalConfigFile.SECTION):
            raise HistoricalConfigFile.RequiredSection(
                HistoricalConfigFile.SECTION)

        for option in [HistoricalConfigFile.BASE_PATH,
                       HistoricalConfigFile.DB_NAME,
                       HistoricalConfigFile.PATCH_PATH,
                       HistoricalConfigFile.SQL_PATH]:
            if self._parser.has_option(HistoricalConfigFile.SECTION,
                                       option):
                continue
            raise HistoricalConfigFile.RequiredOption(option)


    def __get_base_path(self):
        if self._default:
            return self._default.base_path

        return self._parser.get(HistoricalConfigFile.SECTION,
                                HistoricalConfigFile.BASE_PATH)

    def __get_db_name(self):
        if self._default:
            return self._default.db_name

        return self._parser.get(HistoricalConfigFile.SECTION,
                                HistoricalConfigFile.DB_NAME)

    def __get_patch_path(self):
        if self._default:
            return self._default.patch_path

        return self._parser.get(HistoricalConfigFile.SECTION,
                                HistoricalConfigFile.PATCH_PATH)

    def __get_sql_path(self):
        if self._default:
            return self._default.sql_path

        return self._parser.get(HistoricalConfigFile.SECTION,
                                HistoricalConfigFile.SQL_PATH)


    base_path  = property(fget=__get_base_path)
    db_name    = property(fget=__get_db_name)
    patch_path = property(fget=__get_patch_path)
    sql_path   = property(fget=__get_sql_path)


class HistoricalDb(object):

    TABLE_NAME       = 'historical'
    TABLE_DEFINITION = '''
CREATE TABLE historical (
   id INTEGER PRIMARY KEY AUTOINCREMENT,
   host TEXT,
   db_name TEXT,
   patch_file TEXT,
   out_file TEXT,
   date DATE)
'''

    def __init__(self, path='.', db_name='historical.sqlite',
                 patch_path='patches', sql_path='sqls'):
        self._path       = os.path.realpath(os.path.expanduser(path))
        self._patch_path = os.path.realpath(os.path.expanduser(
                '%s%s%s' % (self._path, os.path.sep, patch_path)))
        self._sql_path   = os.path.realpath(os.path.expanduser(
                '%s%s%s' % (self._path, os.path.sep, sql_path)))
        self._db_name    = db_name
        self._connection = sqlite3.connect('%s%s%s' % (self._path,
                                                       os.path.sep,
                                                       self._db_name))
        self._cursor = self._connection.cursor()
        self.__create_table()
        self.__create_directories()

    def __del__(self):
        if self._cursor:
            self._cursor.close()

        if self._connection:
            self._connection.close()

    def __exist_table(self, table):
        query = "SELECT name FROM sqlite_master WHERE " \
            "type='table' AND name='%s';" % table
        self._cursor.execute(query)
        rows = self._cursor.fetchall()
        for row in rows:
            if row[0] == table:
                return True

        return False

    def __create_directories(self):
        for path in [self._sql_path, self._patch_path]:
            if os.path.isdir(path):
                continue
            os.mkdir(path)

    def __create_table(self):
        if self.__exist_table(HistoricalDb.TABLE_NAME):
            return

        self._cursor.execute(HistoricalDb.TABLE_DEFINITION)
        self._connection.commit()

    @staticmethod
    def row_to_historic(row):
        host       = row[1]
        db_name    = row[2]
        patch_file = row[3]
        out_file   = row[4]

        entry = HistoricEntry(db_name, patch_file, host, out_file)
        entry.date   = row[5]
        entry.the_id = row[0]
        entry.the_id = int(row[0])

        return entry

    def add_historic(self, historic):
        if historic.out_file and historic.out_file != '':
            shutil.copy(historic.out_file,
                        '%s%s' % (self._sql_path, os.path.sep))

        shutil.copy(historic.patch_file,
                    '%s%s' % (self._patch_path, os.path.sep))

        insert = "INSERT INTO %s " \
            "(host, db_name, patch_file, out_file, date) " \
            "VALUES (?, ?, ?, ?, ?)" % HistoricalDb.TABLE_NAME

        self._cursor.execute(insert, (
                historic.host,
                historic.db_name,
                os.path.split(historic.patch_file)[1],
                os.path.split(historic.out_file)[1],
                historic.date))
        self._connection.commit()

    def _get_list_from_select(self, select, values=()):
        entries = []
        for row in self._cursor.execute(select, values):
            entries.append(HistoricalDb.row_to_historic(row))

        return entries

    def get_list_all(self):
        select = "SELECT * FROM %s" % HistoricalDb.TABLE_NAME
        return self._get_list_from_select(select)

    def get_list_entries(self, the_ids):
        select = "SELECT * FROM %s WHERE " % HistoricalDb.TABLE_NAME
        for the_id in the_ids[:-1]:
            select += " id = ? OR "
        select += " id = ? "
        return self._get_list_from_select(select, tuple(the_ids))

    def get_list_db(self, command):
        select = "SELECT * FROM %s WHERE db_name = ?" % \
            HistoricalDb.TABLE_NAME
        return self._get_list_from_select(select, (command.arguments[0],))

    def get_list_since(self, command):
        select = "SELECT * FROM %s WHERE date > ?" % HistoricalDb.TABLE_NAME
        return self._get_list_from_select(select, (command.arguments[0],))

    def get_list_until(self, command):
        select = "SELECT * FROM %s WHERE date < ?" % HistoricalDb.TABLE_NAME
        return self._get_list_from_select(select, (command.arguments[0],))

    def get_list_between(self, command):
        select = "SELECT * FROM %s WHERE id > (? - 1) AND id < (? + 1)" % \
            (HistoricalDb.TABLE_NAME)
        return self._get_list_from_select(select, (command.arguments[0],
                                                   command.arguments[1]))

    def get_list_host(self, command):
        select = "SELECT * FROM %s WHERE host = ?" % \
            HistoricalDb.TABLE_NAME
        return self._get_list_from_select(select, (command.arguments[0],))

    def show_entries(self, entries):
        default_sizes = [3, 9, 8, 11, 9, 5]
        sizes = default_sizes[:]
        if not entries:
            print(HistoryFormatter.header(sizes))
            print(HistoryFormatter.line(sizes))
            return

        max_id      = max(entries, key=lambda x : x.the_id)
        max_host    = max(entries, key=lambda x : len(x.host))
        max_db_name = max(entries, key=lambda x : len(x.db_name))
        max_patch   = max(entries, key=lambda x : len(x.patch_file))
        max_out     = max(entries, key=lambda x : len(x.out_file))
        max_date    = max(entries, key=lambda x : len(x.date))

        max_sizes     = [len(str(max_id.the_id))  + 1 ,
                         len(max_host.host)       + 1 ,
                         len(max_db_name.db_name) + 1 ,
                         len(max_patch.patch_file)+ 1 ,
                         len(max_out.out_file)    + 1 ,
                         len(max_date.date)       + 1 ]
        for i in range(len(sizes)):
            if sizes[i] > max_sizes[i]:
                continue
            sizes[i] = max_sizes[i]

        print(HistoryFormatter.header(sizes))
        for entry in entries:
            print(entry.to_string(sizes))
        print(HistoryFormatter.line(sizes))

    def list_all(self, command):
        self.show_entries(self.get_list_all())

    def list_db(self, command):
        self.show_entries(self.get_list_db(command))

    def list_since(self, command):
        self.show_entries(self.get_list_since(command))

    def list_until(self, command):
        self.show_entries(self.get_list_until(command))

    def list_between(self, command):
        self.show_entries(self.get_list_between(command))

    def list_host(self, command):
        self.show_entries(self.get_list_host(command))

    def delete_historic(self, historic):
        if historic.out_file and historic.out_file != '':
            out_file = '%s%s%s' % (self._sql_path,
                                   os.path.sep,
                                   historic.out_file)
            if os.path.isfile(out_file):
                os.remove(out_file)

        patch_file = '%s%s%s' % (self._patch_path,
                                 os.path.sep,
                                 historic.patch_file)
        if os.path.isfile(patch_file):
            os.remove(patch_file)

        self._cursor.execute(
            "DELETE FROM %s WHERE id = ?" % HistoricalDb.TABLE_NAME,
            (historic.the_id,))
        self._connection.commit()

    def delete_all(self, command):
        for entry in self.get_list_all():
            self.delete_historic(entry)

    def delete_entries(self, command):
        for entry in self.get_list_entries(command):
            self.delete_historic(entry)

    def delete_db(self, command):
        for entry in self.get_list_db(command):
            self.delete_historic(entry)

    def delete_since(self, command):
        for entry in self.get_list_since(command):
            self.delete_historic(entry)

    def delete_until(self, command):
        for entry in self.get_list_until(command):
            self.delete_historic(entry)

    def delete_between(self, command):
        for entry in self.get_list_between(command):
            self.delete_historic(entry)

    def delete_host(self, command):
        for entry in self.get_list_host(command):
            self.delete_historic(entry)

    def _get_entry_by_id(self, entry_id):
        self._cursor.execute(
            "SELECT * FROM %s WHERE id = ?" % HistoricalDb.TABLE_NAME,
            (entry_id,))
        rows  = self._cursor.fetchall()

        if len(rows) > 0:
            return HistoricalDb.row_to_historic(rows[0])

        raise Exception('The entry with ID = %s does not exist' % entry_id)


    def __get_out_file_fd(self, entry):
        if entry.out_file and entry.out_file != '':
            return '%s%s%s' % (self._sql_path,
                               os.path.sep,
                               entry.out_file)
        return None

    def __get_patch_file_fd(self, entry):
        return  '%s%s%s' % (self._patch_path,
                            os.path.sep,
                            entry.patch_file)

    def export(self, entries, dst):
        if not os.path.exists(dst):
            raise Exception('The destiny [%s] does not exist' % dst)

        if not  os.path.isdir(dst):
            raise Exception('The destiny [%s] is not a directory' % dst)

        for entry in entries:
            self.__export(entry, dst)

    def __export(self, entry_id, dst):
        out_dir = '%s%s%s' % (dst, os.path.sep, entry_id)
        os.mkdir(out_dir)

        to_export = self._get_entry_by_id(entry_id)
        patch_file_fd = self.__get_patch_file_fd(to_export)
        out_file_fd   = self.__get_out_file_fd(to_export)

        for value, name in [(to_export.db_name, 'db.name'),
                            (to_export.host   , 'host.name'),
                            (to_export.date   , 'export.date')]:
            name_file = '%s%s%s' % (out_dir, os.path.sep, name)
            name_fd   = open(name_file, 'w')
            name_fd.write(value)
            name_fd.close()

        shutil.copy(patch_file_fd, out_dir)
        shutil.copy(out_file_fd, out_dir)

    def see(self, entry_id):
        entry = self._get_entry_by_id(entry_id)

        out_file = ''
        if entry.out_file and entry.out_file != '':
            out_file_fd = self.__get_out_file_fd(entry)
            if os.path.isfile(out_file_fd):
                out_file = open(out_file_fd, 'r').read()

        patch_file = ''
        patch_file_fd = self.__get_patch_file_fd(entry)
        if os.path.isfile(patch_file_fd):
            patch_file = open(patch_file_fd, 'r').read()

        pf_header = '%s : %s' % (StrColor.paint('Patch file', Color.WHITE,
                                                Format.BOLD),
                                 StrColor.paint(entry.patch_file, Color.WHITE,
                                                Format.BOLD))
        of_header = '%s : %s' % (StrColor.paint('Generated file', Color.WHITE,
                                                Format.BOLD),
                                 StrColor.paint(entry.out_file, Color.WHITE,
                                                Format.BOLD))

        sep_blue = StrColor.paint('-'*80, Color.BLUE, Format.BOLD,
                                  Intensity.TEXT_STRONG_INSTENSITY)
        sep_gray = StrColor.paint('-'*80, Color.GRAY, Format.BOLD,
                                  Intensity.TEXT_STRONG_INSTENSITY)

        print '%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s' % \
            (sep_gray, pf_header, sep_blue, patch_file,
             sep_gray, of_header, sep_blue, out_file)


class HistoryUI(object):

    def __init__(self, config_file):
        self._config = HistoricalConfigFile(config_file)
        self._historical_db = HistoricalDb(
            self._config.base_path,
            self._config.db_name,
            self._config.patch_path,
            self._config.sql_path)

    def execute(self, raw_cmd):
        cmd = HistoryCommandParser.parse(raw_cmd)
        if cmd.is_list_command():
            getattr(self._historical_db, 'list_%s' % cmd.subcommand)(cmd)
        elif cmd.is_see_command():
            self._historical_db.see(cmd.arguments[0])
        elif cmd.is_export_command():
            n_args = len(cmd.arguments)
            self._historical_db.export(cmd.arguments[0:n_args - 1],
                                       cmd.arguments[n_args   - 1])
        elif cmd.is_delete_command():
            getattr(self._historical_db, 'delete_%s' % cmd.subcommand)(cmd)

