'''
    db-patch applies static and dynamic patches to databases using ODBC or
    native connectors.
    Copyright (C) 2012, 2013  Miguel A. Martinez Pinedo

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
    02110-1301  USA
'''


import os
import os.path
import sys
import imp
import inspect


def load_module(full_name):
    '''Load a module using the full path passed by parameter'''
    (path, name) = os.path.split(full_name)
    name = os.path.splitext(name)[0]

    (fdesc, filename, data) = imp.find_module(name, [path])
    return imp.load_module(name, fdesc, filename, data)

def get_classes(module):
    '''
    Recover all the classes of one module passed by parameter.
    The module object must be obtained using 'load_module' method.
    '''
    return inspect.getmembers(module, inspect.isclass)

def new_class(module, class_name, **kargs):
    '''
    Create a class instance.

    module - The module (obtained using 'load_module' method)
    class_name - The class name to instantiate
    kargs - The arguments used for instantiate the class
    '''
    the_class = getattr(module, class_name)
    return the_class(**kargs)

def new_class_child_of(module, parent_class, **kargs):
    '''
    Create a class instance of the child class of the class passed
    by parameter.

    module - The module (obtained using 'load_module' method)
    parent_class - The parent class name
    kargs - The arguments used for instantiate the class
    '''
    loaded_module = load_module(module)
    the_classes   = get_classes(loaded_module)
    hierarchies   = []
    for name, loaded_class in the_classes:
        hierarchies.append((name, inspect.getmro(loaded_class)))

    for name, hierarchy in hierarchies:
        if len(hierarchy) < 2:
            continue

        for child in hierarchy:
            if str(child) != parent_class:
                continue

            return getattr(loaded_module, name)(**kargs)
    
    return None


class Registry(object):
    '''
    Class that resgister classes. The purpose of this class is to be
    used as plugin engine.

    The plugins register its classes in the Registry, and the factories
    instantiate the registred classes.
    '''

    REGISTRY = {}

    @classmethod
    def register(cls, module, the_class):
        '''
        Register the class using the module and the class atributte
        REGISTRY_ID as indexes.
        '''
        if module not in Registry.REGISTRY.keys():
            Registry.REGISTRY[module] = {}

        Registry.REGISTRY[module][the_class.REGISTRY_ID] = the_class

    @classmethod
    def get(cls, module, the_class_name):
        '''Recover the class using the module and the class name'''
        return Registry.REGISTRY[module][the_class_name]



class Loader(object):
    '''
    Base class for create plugins factories. This class find python
    files in the path passed in the constructor, and register them
    calling to the 'register' module method.

    The classes are registered in the dbpatch.loader.Registry using
    the module_name passed by parameter at init time.

    The child classes MUST implement the 'log' method.
    '''

    def __init__(self, module_name, base_path, extra_path=None):
        '''
        Contructor. Load all the clases of the directories base_path
        and extra_path, registering in the Registry using the
        module_name.
        '''
        self._paths = [base_path]
        self._module_name = module_name

        if extra_path:
            full_path = os.path.realpath(os.path.expanduser(extra_path))
            if os.path.isdir(full_path):
                self._paths.append(extra_path)
                self.log().trace('Using extra path [%s]' % full_path)

        self.__update_sys_path()
        self.__load_class()
        self.__restore_sys_path()

    def __update_sys_path(self):
        '''
        Update the sys.path with the paths that contains the python
        files
        '''
        for new_path in self._paths:
            if new_path in sys.path:
                continue
            self.log().trace('Adding [%s] to sys.path' % new_path)
            sys.path.append(new_path)

    def __restore_sys_path(self):
        '''
        Restore the sys.path (remove the paths added with
        __update_sys_path)
        '''
        for new_path in self._paths:
            if new_path not in sys.path:
                continue
            self.log().trace('Removing [%s] of sys.path' % new_path)
            sys.path.remove(new_path)

    def __recover_class_files(self):
        '''Return a list with the python files defined in the paths'''
        class_files = []

        for class_path in self._paths:
            if not os.path.isdir(class_path):
                continue

            for name in os.listdir(class_path):
                if name.startswith('.') or \
                       name.startswith('#') or \
                       name.startswith('__') or \
                       not name.endswith('.py') :
                    continue

                class_files.append(name)

        return class_files

    def __load_class(self):
        '''Register the classes defined in the python files'''
        class_files = self.__recover_class_files()

        for the_file in class_files:
            module = __import__(the_file[:-3])
            try:
                module.register(self._module_name)
            except Exception, error:
                self.log().error('Required register() method!! in '
                                 '%s [%s]' % (the_file, error))

    def create(self, class_identifier, **args):
        '''Instantiate a class'''
        self.log().trace('Creating [%s.%s]' % (self._module_name,
                                               class_identifier))
        return Registry.get(self._module_name, class_identifier)(**args)

    def log(self):
        '''Return a logger'''
        raise NotImplementedError



