'''
    db-patch applies static and dynamic patches to databases using ODBC or
    native connectors.
    Copyright (C) 2012, 2013  Miguel A. Martinez Pinedo

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
    02110-1301 USA
'''

import os
import sys
import time
import logging

import dbpatch.options
from dbpatch.singleton import Singleton
from dbpatch.context import Context


class TraceableLog(logging.getLoggerClass()):
    '''
    Extends the Log class of logging module adding a new level:
    TRACE
    '''

    TRACE_LEVEL = 5
    TRACE_TAG   = 'TRACE'

    def trace(self, msg, *args, **kwargs):
        if not self.isEnabledFor(TraceableLog.TRACE_LEVEL):
            return
        self._log(TraceableLog.TRACE_LEVEL, msg, args, **kwargs)



class LogFactory(object):
    '''
    Singleton used for create loggers taking into account the command
    line options passed to db-patch
    '''

    __metaclass__ = Singleton

    SQL_PREFIX = 'SQL'

    def __init__(self):
        self._options = Context().options
        self._loggers = {}
        self._customize_logging()
        self._level = self._set_log_level()
        self._log   = self.get_logger(LogFactory.__name__)

    def _customize_logging(self):
        for handler in logging.Logger.root.handlers:
            logging.Logger.root.removeHandler(handler)

        logging.addLevelName(TraceableLog.TRACE_LEVEL,
                             TraceableLog.TRACE_TAG)
        logging.setLoggerClass(TraceableLog)

    def _set_log_level(self):
        level = self._options.get_option(dbpatch.options.VERBOSE)

        if level == 1 :
            return logging.DEBUG
        elif level == 2 :
            return TraceableLog.TRACE_LEVEL
        else:
            return logging.INFO

    def _configure_log(self, log):
        log.addHandler(ConsoleLog.get_handler(self._options))

    @staticmethod
    def configure_sql_log(log=None, sql_log_prefix='SQL'):
        '''
        Configure the log used for show and store the SQL generated.
        '''
        if not Context().current_db:
            return

        if not log:
            log = logging.getLogger(sql_log_prefix)

        for handler in log.handlers:
            log.removeHandler(handler)

        if not Context().options.get_option(dbpatch.options.SQL_CLEAN):
            file_handler = SqlLog.get_file_handler()
        else:
            file_handler = SqlLog.get_tmp_file_handler()

        file_handler.setLevel(logging.NOTSET)
        log.addHandler(file_handler)

        if Context().options.get_option(dbpatch.options.SQL_SHOW) and \
                not Context().showing_sql:
            console_handler = SqlLog.get_console_handler()
            log.addHandler(console_handler)
            Context().showing_sql = True

    def get_logger(self, log_name):
        '''
        Create a logger taking into account the program options and
        using the 'log_name'. If the log is already created it will
        be returned.
        '''
        if log_name in self._loggers.keys():
            return self._loggers[log_name]

        log = logging.getLogger(log_name)
        for handler in log.handlers:
            log.removeHandler(handler)

        if log_name.upper().startswith(LogFactory.SQL_PREFIX):
            LogFactory.configure_sql_log(log)
        else:
            self._configure_log(log)

        log.setLevel(self._level)

        self._loggers[log_name] = log

        return log



class ConsoleLog(object):

    @staticmethod
    def get_handler(level):
        if level == logging.DEBUG or logging == TraceableLog.TRACE_LEVEL:
            return ConsoleLog.get_debug_handler()

        return ConsoleLog.get_info_handler()

    @staticmethod
    def get_info_handler():
        handler   = logging.StreamHandler(stream=sys.stdout)
        formatter = logging.Formatter('%(levelname)s| %(message)s')
        handler.setFormatter(formatter)

        return handler
    
    @staticmethod
    def get_debug_handler():
        handler = logging.StreamHandler(stream=sys.stdout)
        formatter = logging.Formatter(
            '%(asctime)s|%(levelname)s| %(message)s',
            '%m/%d/%Y %I:%M:%S')

        handler.setFormatter(formatter)

        return handler
    


class SqlLog(object):

    @staticmethod
    def create_file_handler(name_file):
        Context().current_sql_log = name_file
        handler = logging.FileHandler(name_file, 'w')
        Context().log_files[Context().current_db] = name_file
        handler.setFormatter(logging.Formatter('%(message)s'))
        return handler

    @staticmethod
    def get_tmp_file_handler():
        import uuid
        name_file = 'db-patch-%s' % uuid.uuid1()
        Context().current_sql_log = name_file
        Context().log_files[Context().current_db] = name_file
        return SqlLog.create_file_handler(name_file)

    @staticmethod
    def get_file_handler():
        name_file = SqlLog.get_output_file()
        return SqlLog.create_file_handler(name_file)

    @staticmethod
    def get_console_handler():
        handler = logging.StreamHandler(stream=sys.stdout)
        handler.setFormatter(logging.Formatter('SQL|%(message)s'))

        return handler

    @staticmethod
    def build_output_file_name(host=None):
        patch_file = Context().options.get_option(dbpatch.options.PATCH_FILE)
        file_name  = os.path.split(patch_file)[1]
        file_name  = os.path.splitext(file_name)[0]

        the_host = host
        if not host:
            the_host = Context().options.get_option(dbpatch.options.DB_HOST)

        time_stamp  = time.strftime('%m_%d_%Y__%H_%M_%S')
        name_format = Context().options.get_option(dbpatch.options.SQL_GEN_FILE)
        
        name = name_format.replace('NAME', file_name)
        name = name.replace('TIME', time_stamp)
        name = name.replace('HOST', the_host)

        if Context().current_db:
            return Context().current_db + '_' + name

        return name

    @staticmethod
    def get_output_file(host=None):
        file_name  = SqlLog.build_output_file_name(host)
        output_dir = Context().options.get_option(dbpatch.options.SQL_GEN_DIR)
        if not os.path.isdir(output_dir):
            os.mkdir(output_dir)

        return '%s/%s' % (output_dir, file_name)

