'''
    db-patch applies static and dynamic patches to databases using ODBC or
    native connectors.
    Copyright (C) 2012, 2013  Miguel A. Martinez Pinedo

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301
    USA
'''


import sys

import dbpatch.optwrapper
from dbpatch.context import Context
from dbpatch.color import *


# Options for ODBC connections
ODBC_DRIVER     = 'odbc_driver'
ODBC_DSN        = 'odbc_dsn'
ODBC_EXTRA      = 'odbc_extra'
ODBC_FULL_CONN  = 'odbc_full_conn'
ODBC_INST_INI   = 'odbc_instini'
ODBC_PLUGINS    = 'odbc_plugins'
CONN_PLUGINS    = 'connection_plugins'
CONN_NATIVE     = 'native'

# Options for database
DB_USER    = 'db_user'
DB_PASS    = 'db_pass'
DB_HOST    = 'db_host'
DB_PORT    = 'db_port'
DB_ENGINE  = 'db_engine'
DB_SOCKET  = 'db_socket'
DB_CHARSET = 'db_charset'
DB_CHARSET_POLICY = 'db_charset_policy'

# Options for filter the database list
EXCLUDE   = 'exclude'
INCLUDE   = 'include'
DATABASES = 'databases'

# Patch options
PATCH_FILE    = 'patch_file'
PATCH_TYPE    = 'patch_type'
PATCH_PLUGINS = 'patch_plugins'

# Generated SQL options
SQL_CLEAN    = 'clean'
SQL_GEN_DIR  = 'generated_dir'
SQL_GEN_FILE = 'gen_file'
SQL_SHOW     = 'show_sql'

# Run options
DRY_RUN  = 'dry_run'
VERBOSE  = 'verbose'
PRE_CMD  = 'pre_cmd'
POST_CMD = 'post_cmd'

# Misc options
EXTRA_OPTS = 'extra_opts'
ARGS_FILE  = 'args_file'
HISTORICAL_FILE    = 'historical_file'
HISTORICAL_ENABLED = 'historical_enabled'



class OdbcOptions(dbpatch.optwrapper.OptionGroupWrapper):

    ODBC_TITLE = 'Connection and ODBC options'

    ODBC_OPTS = [
        dbpatch.optwrapper.LongOption('--odbc-driver',
                   'Define the ODBC driver to use. '
                   'It must be defined in the odbcinst.ini',
                   None,
                   ODBC_DRIVER),
        dbpatch.optwrapper.LongOption('--odbc-dsn',
                   'Define a DSN for ODBC connections',
                   None,
                   ODBC_DSN),
        dbpatch.optwrapper.LongOption('--odbc-extra',
                   'Provide extra options for make the ODBC '
                   'connection string',
                   None,
                   ODBC_EXTRA),
        dbpatch.optwrapper.LongOption('--odbc-full-connection',
                   'Define the full ODBC connection string',
                   None,
                   ODBC_FULL_CONN),
        dbpatch.optwrapper.LongOption('--odbc-inst-ini',
                   'Full path to odbcinst.ini. '
                   'By default: /etc/odbcinst.ini',
                   '/etc/odbcinst.ini',
                   ODBC_INST_INI),
        dbpatch.optwrapper.BooleanOption('-N', '--native',
                      'Use a native driver for database connection. '
                      'By default uses ODBC (requires to configure '
                      'the odbcinst.ini file)',
                      False,
                      CONN_NATIVE,
                      'store_true'),
        dbpatch.optwrapper.LongOption('--odbc-plugins',
                   'Define the ODBC query builder path '
                   'for locate custom ODBC connections.'
                   'By default: ~/.dbpatch/odbc',
                   '~/.dbpatch/odbc',
                   ODBC_PLUGINS),
        dbpatch.optwrapper.LongOption('--connection-plugins',
                   'Extra path for locate custom native connections. '
                   'By default: ~/.dbpatch/connection',
                   '~/.dbpatch/connection',
                   CONN_PLUGINS),

        ]

    def __init__(self, parser):
        dbpatch.optwrapper.OptionGroupWrapper.__init__(self, parser)

    def get_title(self):
        return OdbcOptions.ODBC_TITLE

    def get_description(self):
        return ''

    def get_help(self):
        return OdbcOptions.ODBC_TITLE

    def get_raw_options(self):
        return OdbcOptions.ODBC_OPTS[:]



class DatabaseOptions(dbpatch.optwrapper.OptionGroupWrapper):

    DB_TITLE = 'Database options'

    DB_OPTS = [
        dbpatch.optwrapper.ShortLongOption('-u', '--db-user',
                        'Database user. Mandatory',
                        None,
                        DB_USER),
        dbpatch.optwrapper.ShortLongOption('-p', '--db-pass',
                        'Database password. Mandatory',
                        None,
                        DB_PASS),
        dbpatch.optwrapper.ShortLongOption('-H', '--db-host',
                        'Database host. By default "localhost"',
                        'localhost',
                        DB_HOST),
        dbpatch.optwrapper.ShortLongOption('-P', '--db-port',
                        'Database port',
                        None,
                        DB_PORT),
        dbpatch.optwrapper.ShortLongOption('-E', '--db-engine',
                        'Database engine (MySQL, PostgreSQL, etc). Mandatory',
                        None,
                        DB_ENGINE),
        dbpatch.optwrapper.ShortLongOption('-C', '--db-charset',
                        'The charset used for database connections.'\
                        'By default: ascii',
                        'ascii',
                        DB_CHARSET),
        dbpatch.optwrapper.ShortLongOption('-s', '--db-socket',
                        'Database Unix socket',
                        None,
                        DB_SOCKET),
        dbpatch.optwrapper.ShortLongOption('-e', '--exclude',
                        'Regular expression for exclude databases.'
                        'Mandatory to use python regular expressions (re module). '
                        'Example: -e "a[a-z]+"',
                        None,
                        EXCLUDE),
        dbpatch.optwrapper.ShortLongOption('-i', '--include',
                        'Regular expression for include databases.'
                        'See python re module',
                        None,
                        INCLUDE),
        dbpatch.optwrapper.ShortLongOption('-d', '--databases',
                        'List of databases to patch.'
                        'Examples: -d W24_WE1 or -d W24_WE1,W24_WH6',
                        None,
                        DATABASES),
        dbpatch.optwrapper.LongOption('--db-charset-policy',
                   'The policy to apply when the charset can not full replace strings. '
                   'Posible values: [ignore|replace|xmlcharrefreplace]. '
                   'By default: ignore',
                   'ignore',
                   DB_CHARSET_POLICY)
        ]

    def __init__(self, parser):
        dbpatch.optwrapper.OptionGroupWrapper.__init__(self, parser)

    def get_title(self):
        return DatabaseOptions.DB_TITLE

    def get_description(self):
        return '' #DatabaseOptions.DB_DESC

    def get_help(self):
        return DatabaseOptions.DB_TITLE

    def get_raw_options(self):
        return DatabaseOptions.DB_OPTS[:]


class PatchOptions(dbpatch.optwrapper.OptionGroupWrapper):

    PATCH_TITLE = 'Patch options'

    PATCH_OPTS = [
        dbpatch.optwrapper.ShortLongOption(
            '-f', '--file',
            'File that contains the patch. The patch executor uses the extension for '
            'runs each patch. For the ".sql" patches the mydb keyword will be '
            'replaced by each database name. Mandatory',
            None,
            PATCH_FILE),
        dbpatch.optwrapper.LongOption('--patch-plugins',
                   'Extra path for locate customized patchs. '
                   'By default: ~/.dbpatch/patch',
                   '~/.dbpatch/patch',
                   PATCH_PLUGINS),
        dbpatch.optwrapper.LongOption('--patch-type',
                   'Patch type. Supported: sql, python. '
                   'Extensible via patch plugins (see --patch-plugins).'
                   'By default the patch type will be determined by the file extension',
                   None,
                   PATCH_TYPE),
        ]

    def __init__(self, parser):
        dbpatch.optwrapper.OptionGroupWrapper.__init__(self, parser)

    def get_title(self):
        return PatchOptions.PATCH_TITLE

    def get_description(self):
        return ''

    def get_help(self):
        return PatchOptions.PATCH_TITLE

    def get_raw_options(self):
        return PatchOptions.PATCH_OPTS[:]


class SqlOptions(dbpatch.optwrapper.OptionGroupWrapper):

    SQL_TITLE = 'SQL tracing options'

    SQL_OPTS = [
        dbpatch.optwrapper.BooleanOption('-c', '--clean',
                      'Avoid that all the SQL executed will be stored',
                      False,
                      SQL_CLEAN,
                      'store_true'),
        dbpatch.optwrapper.LongOption('--gen-dir',
                   'Directory where the SQL sentences for each patch will be stored. '
                   'By default: sql_sentences',
                   'sql_sentences',
                   SQL_GEN_DIR),
        dbpatch.optwrapper.LongOption('--gen-file',
                   'Determines the name of the file where the generated SQL is stored.'
                   'The "HOST" flag will be replaced by the hostname.'
                   'The "NAME" flag will be replaced by the patch name without extension.'
                   'The "TIME" flag will be replaced by a timestamp. '
                   'By default: "HOST_NAME_TIME.sql"',
                   'HOST_NAME_TIME.sql',
                   SQL_GEN_FILE),
        dbpatch.optwrapper.BooleanOption('-S', '--show-sql',
                      'Show the generated SQL sentences in the output',
                      False,
                      SQL_SHOW,
                      'store_true'),
        ]

    def __init__(self, parser):
        dbpatch.optwrapper.OptionGroupWrapper.__init__(self, parser)

    def get_title(self):
        return SqlOptions.SQL_TITLE

    def get_description(self):
        return ''

    def get_help(self):
        return SqlOptions.SQL_TITLE

    def get_raw_options(self):
        return SqlOptions.SQL_OPTS[:]


class RunOptions(dbpatch.optwrapper.OptionGroupWrapper):

    RUN_TITLE = 'Run options'

    RUN_OPTS = [
        dbpatch.optwrapper.BooleanOption('-t', '--dry-run',
                      'Generate all the SQL sentences without modify the database',
                      False,
                      DRY_RUN,
                      'store_true'),
        dbpatch.optwrapper.BooleanOption('-v', '--verbose',
                      'Increase the verbose level', 0, VERBOSE, 'count'),
        dbpatch.optwrapper.LongOption('--pre-cmd',
                   "Command to be executed before of execute the patch. "
                   "By example --pre-cmd 'find -name \"*.pyc\" -delete' ",
                   None,
                   PRE_CMD),
        dbpatch.optwrapper.LongOption('--post-cmd',
                   "Command to be executed after of execute the patch. "
                   "By example --post-cmd "
                   "'tar -czf applied.tgz sql_generated/; "
                   "rm -rf sql_generated/' ",
                   None,
                   POST_CMD)
        ]

    def __init__(self, parser):
        dbpatch.optwrapper.OptionGroupWrapper.__init__(self, parser)

    def get_title(self):
        return RunOptions.RUN_TITLE

    def get_description(self):
        return ''

    def get_help(self):
        return RunOptions.RUN_TITLE

    def get_raw_options(self):
        return RunOptions.RUN_OPTS[:]


class MiscOptions(dbpatch.optwrapper.OptionGroupWrapper):

    MISC_TITLE = 'Extra options'

    MISC_OPTS = [
        dbpatch.optwrapper.LongOption('--extra-opts',
                   'Allow use a configuration file for introduce new '
                   'options. The options will be stored as '
                   '"section_name.option_name"',
                   None,
                   EXTRA_OPTS),
        dbpatch.optwrapper.LongOption('--args-file',
                   'Allow use a file for introduce the arguments.'
                   'The command line options will overwrite the options '
                   'defined in the file',
                   None,
                   ARGS_FILE),
        dbpatch.optwrapper.BooleanOption(None, '--enable-historic',
                      'Store the patch and the SQL senteces generated'
                      ' in the historical database',
                      False,
                      HISTORICAL_ENABLED,
                      'store_true'),
        dbpatch.optwrapper.LongOption('--history-cfg',
                   'Allow define the configuration file used for store '
                   'the executed patches. ',
                   '',
                   HISTORICAL_FILE),
        ]

    def __init__(self, parser):
        dbpatch.optwrapper.OptionGroupWrapper.__init__(self, parser)

    def get_title(self):
        return MiscOptions.MISC_TITLE

    def get_description(self):
        return ''

    def get_help(self):
        return MiscOptions.MISC_TITLE

    def get_raw_options(self):
        return MiscOptions.MISC_OPTS[:]


class DbPatchOptions(dbpatch.optwrapper.Options):
    def __init__(self, parser, groups, mandatories):
        dbpatch.optwrapper.Options.__init__(self, parser, groups,
                                            mandatories)

        self._databases = {'include'   : [],
                           'exclude'   : [],
                           'databases' : []}

    def parse(self):
        options = dbpatch.optwrapper.Options.parse(self)

        args_file = getattr(options, ARGS_FILE)
        if args_file:
            args = open(args_file).read().split()
            arguments = sys.argv[1:]
            arguments.extend(args)
            options, args = self._parser.parse_arguments(arguments)

        for name in [INCLUDE, EXCLUDE, DATABASES]:
            if not getattr(options, name):
                continue
            self._databases[name] = getattr(options,name).split(',')

        if self._options[EXTRA_OPTS]:
            self.set_options_from_file(self._options[EXTRA_OPTS])

        Context().options = self

    def get_include_dbs(self):
        return self._databases[INCLUDE]

    def get_exclude_dbs(self):
        return self._databases[EXCLUDE]

    def get_only_dbs(self):
        return self._databases[DATABASES]

