'''
    db-patch applies static and dynamic patches to databases using ODBC or
    native connectors.
    Copyright (C) 2012, 2013  Miguel A. Martinez Pinedo

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301
    USA
'''

import sys

import dbpatch.color


class ConfigurationFileException(Exception):
    def __init__(self, message):
        Exception.__init__(self, message)


class RequiredOption(Exception):
    def __init__(self, message):
        Exception.__init__(self, message)


class Option(object):

    SHORT_OPTS = 'short_opts'
    LONG_OPTS = 'long_opts'
    DESTINY = 'destiny'
    HELP = 'help_msg'
    DEFAULT = 'default_value'
    BEHAVIOUR = 'behaviour'

    def __init__(self):
        self._help = None
        self._description = None
        self._short_opts = None
        self._long_opts = None
        self._default_value = None
        self._destiny = None

    def add(self, option_parse_wrapper):
        raise NotImplementedError

    def __get_help(self):
        return self._help

    def __set_help(self, the_help):
        self._help = the_help

    def __set_short_opts(self, short_opts):
        self._short_opts = short_opts

    def __get_short_opts(self):
        return self._short_opts

    def __set_long_opts(self, long_opts):
        self._long_opts = long_opts

    def __get_long_opts(self):
        return self._long_opts

    def __set_default_value(self, default_value):
        self._default_value = default_value

    def __get_default_value(self):
        return self._default_value

    def __set_destiny(self, destiny):
        self._destiny = destiny

    def __get_destiny(self):
        return self._destiny

    def __set_description(self, description):
        self._description = description

    def __get_description(self):
        return self._description

    help_msg    = property(fget=__get_help, fset=__set_help)
    description = property(fget=__get_description, fset=__set_description)
    short_opts  = property(fget=__get_short_opts, fset=__set_short_opts)
    long_opts   = property(fget=__get_long_opts, fset=__set_long_opts)
    default_val = property(fget=__get_default_value,
                           fset=__set_default_value)
    destiny = property(fget=__get_destiny, fset=__set_destiny)


class LongOption(Option):

    def __init__(self, long_opts, help_msg, default_value, destiny):
        Option.__init__(self)
        self.help_msg = help_msg
        self.long_opts = long_opts
        self.default_value = default_value
        self.destiny = destiny

    def add(self, parser):
        kargs = {Option.LONG_OPTS : self.long_opts,
                 Option.HELP : self.help_msg,
                 Option.DESTINY : self.destiny,
                 Option.DEFAULT : self.default_value
                 }
        parser.add_option(**kargs)
        parser.add_raw_option(self)


class ShortLongOption(Option):

    def __init__(self, short_opts, long_opts, help_msg, default_value,
                 destiny):
        Option.__init__(self)
        self.short_opts = short_opts
        self.long_opts = long_opts
        self.help_msg = help_msg
        self.default_val = default_value
        self.destiny = destiny

    def add(self, parser):
        kargs = {Option.SHORT_OPTS : self.short_opts,
                 Option.LONG_OPTS : self.long_opts,
                 Option.HELP : self.help_msg,
                 Option.DESTINY : self.destiny,
                 Option.DEFAULT : self.default_val
                 }
        parser.add_option(**kargs)
        parser.add_raw_option(self)


class BooleanOption(Option):

    def __init__(self, short_opts, long_opts, help_msg, default_value,
                 destiny, behaviour):
        Option.__init__(self)
        self.short_opts = short_opts
        self.long_opts = long_opts
        self.help_msg = help_msg
        self.default_value = default_value
        self.destiny = destiny
        self.behaviour = behaviour

    def add(self, parser):
        kargs = {Option.SHORT_OPTS : self.short_opts,
                 Option.LONG_OPTS : self.long_opts,
                 Option.HELP : self.help_msg,
                 Option.DESTINY : self.destiny,
                 Option.DEFAULT : self.default_value,
                 Option.BEHAVIOUR : self.behaviour,
                 }
        parser.add_option(**kargs)
        parser.add_raw_option(self)


class OptionParserFactory(object):

    @staticmethod
    def create(the_description):
        info = sys.version_info
        if info.major >= 2 and info.minor >= 7:
            import argparse
            parser = argparse.ArgumentParser(description=the_description)
            return ArgParserWrapper(parser)
        else:
            import optparse
            parser = optparse.OptionParser(description=the_description)
            return OptParserWrapper(parser)


class OptionParserWrapper(object):

    DEST = 'dest'
    HELP = 'help'
    DEFAULT = 'default'
    ACTION = 'action'

    def __init__(self, parser):
        self._parser = parser
        self._raw_options = []

    def add_raw_option(self, raw_option):
        self._raw_options.append(raw_option)

    def add_option(self, **kargs):
        raise NotImplementedError

    def add_group(self, **kargs):
        raise NotImplementedError

    def get_options(self):
        return self._raw_options

    def parse_options(self):
        raise NotImplementedError

    def parse_arguments(self, arguments_line):
        raise NotImplementedError


class ArgumentBuilder(object):

    @staticmethod
    def build_args(**kargs):
        args = []
        if Option.SHORT_OPTS in kargs.keys():
            if kargs[Option.SHORT_OPTS]:
                args.append(kargs[Option.SHORT_OPTS])
        if Option.LONG_OPTS in kargs.keys():
            if kargs[Option.LONG_OPTS]:
                args.append(kargs[Option.LONG_OPTS])

        return args

    @staticmethod
    def build_kargs(**kargs):
        k_args = { OptionParserWrapper.DEST : kargs[Option.DESTINY] ,
                   OptionParserWrapper.HELP : kargs[Option.HELP],
                   }

        optionals = [(Option.DEFAULT, OptionParserWrapper.DEFAULT),
                     (Option.BEHAVIOUR, OptionParserWrapper.ACTION)]

        for optional, key in optionals:
            if optional not in kargs.keys():
                continue
            k_args[key] = kargs[optional]

        return k_args


class ArgParserWrapper(OptionParserWrapper):

    def __init__(self, parser):
        OptionParserWrapper.__init__(self, parser)
        self._parser = parser

    def add_option(self, **kargs):
        arguments = ArgumentBuilder.build_args(**kargs)
        k_args = ArgumentBuilder.build_kargs(**kargs)

        self._parser.add_argument(*arguments, **k_args)

    def add_group(self, group):
        class SubGroup(object):
            def __init__(self, parser, subparser):
                self._subparser = subparser
                self._parser = parser

            def add_option(self, **kargs):
                arguments = ArgumentBuilder.build_args(**kargs)
                k_args = ArgumentBuilder.build_kargs(**kargs)
                self._subparser.add_argument(*arguments, **k_args)

            def add_raw_option(self, raw_opt):
                self._parser.add_raw_option(raw_opt)

        subparser = self._parser.add_argument_group(
            group.get_title(), group.get_description())

        subgroup = SubGroup(self, subparser)

        for option in group.get_raw_options():
            option.add(subgroup)
            subgroup.add_raw_option(option)

    def parse_options(self):
        return self._parser.parse_args()

    def parse_arguments(self, arguments_line):
        return self._parser.parse_args(arguments_line.split())


class OptParserWrapper(OptionParserWrapper):

    def __init__(self, parser):
        OptionParserWrapper.__init__(self, parser)
        self._parser = parser

    def add_option(self, **kargs):
        arguments = ArgumentBuilder.build_args(**kargs)
        k_args = ArgumentBuilder.build_kargs(**kargs)

        self._parser.add_option(*arguments, **k_args)

    def add_group(self, group):
        import optparse
        opt_group = optparse.OptionGroup(self._parser,
                                         group.get_title(),
                                         group.get_description())
        class SubGroup(object):
            def __init__(self, parser, opt_group):
                self._parser = parser
                self._opt_group = opt_group

            def add_option(self, **kargs):
                arguments = ArgumentBuilder.build_args(**kargs)
                k_args = ArgumentBuilder.build_kargs(**kargs)
                self._opt_group.add_option(*arguments, **k_args)

            def add_raw_option(self, raw_opt):
                self._parser.add_raw_option(raw_opt)

        subgroup = SubGroup(self, opt_group)
        for option in group.get_raw_options():
            option.add(subgroup)

        self._parser.add_option_group(opt_group)

    def parse_options(self):
        options, args = self._parser.parse_args()
        return options

    def parse_arguments(self, arguments_line):
        options, args = self._parser.parse_args(arguments_line)
        return options


class OptionGroupWrapper(object):

    def __init__(self, parser):
        self._parser = parser
        self._group  = None

    def get_title(self):
        raise NotImplementedError

    def get_description(self):
        raise NotImplementedError

    def get_help(self):
        raise NotImplementedError

    def get_raw_options(self):
        raise NotImplementedError

    def add_group(self, parser):
        parser.add_group(group=self)


class Options(object):

    MANDATORY_MSG = ' This command requires several options for work.' \
      '\n See the complete option list using -h or --help.'

    MANDATORY_OPTS_MSG = ' The next options are mandatories %s'

    def __init__(self, parser, groups, mandatories):
        self._parser = parser
        self._options = {}
        self._groups = groups
        self._mandatories = mandatories

        for opt_group in self._groups :
            self._parser.add_group(opt_group)

    @staticmethod
    def check_mandatory_options(options, mandatories):
        for mandatory in mandatories:
            if getattr(options, mandatory):
                continue
            fail_red = dbpatch.color.StrColor.paint(
                'FAIL!',
                dbpatch.color.Color.RED,
                dbpatch.color.Format.BOLD)
            print(fail_red + Options.MANDATORY_MSG)
            print(Options.MANDATORY_OPTS_MSG % str(mandatories))
            sys.exit(1)

    def parse(self):
        options = self._parser.parse_options()
        for opt in self._parser.get_options():
            self._options[opt.destiny] = getattr(options, opt.destiny)

        self.check_mandatory_options(options, self._mandatories)

        for group in self._groups:
            for opt in group.get_raw_options():
                self._options[opt.destiny] = getattr(options, opt.destiny)

        return options

    def get_options(self):
        return self._options.copy()

    def get_option(self, the_option):
        return self._options[the_option]

    def set_options(self, options):
        self._options.update(options)

    def get_mandatory_option(self, the_option):
        if not the_option in self._options.keys():
            raise RequiredOption(
                'The option [%s] is mandatory' % the_option)

        value = self._options[the_option]

        if not value:
            raise RequiredOption('The option [%s] is mandatory. '
                                 'The value cannot be None' % the_option)

        return value

    def set_options_from_file(self, the_file):
        import os
        if not os.path.isfile(the_file):
            raise ConfigurationFileException(
                'The file [%s] does not exist' % the_file)

        from ConfigParser import ConfigParser
        config = ConfigParser()
        config.read(the_file)

        for section in config.sections():
            for opt in config.options(section):
                key = '%s.%s' % (section, opt)
                self._options[key] = config.get(section, opt)

