'''
    db-patch applies static and dynamic patches to databases using ODBC or
    native connectors.
    Copyright (C) 2012, 2013  Miguel A. Martinez Pinedo

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301  USA
'''


import re
from dbpatch.log import LogFactory
from dbpatch.context import Context


class StatementBuilder(object):

    DRY_RUN = 'dry_run'

    def __init__(self):
        options = Context().options.get_options()

        if StatementBuilder.DRY_RUN not in options.keys():
            raise dbpatch.options.RequiredOption(StatementBuilder.DRY_RUN)

        self._dry_run = options[StatementBuilder.DRY_RUN]

    def create(self, connection, marked_query, values=[], autocommit=True):
        if self._dry_run:
            return PrepareStatementBuilderDryRun.create(connection,
                                                        marked_query,
                                                        values,
                                                        autocommit)
        return PrepareStatementBuilder.create(connection,
                                              marked_query,
                                              values,
                                              autocommit)


class MissingValues(Exception):
    def __init__(self, message):
        Exception.__init__(self, message)


class PrepareStatementBuilder(object):

    @staticmethod
    def is_select(query):
        select_pattern = re.compile('[\s*]*(SELECT|select)[\s*\w*]*')
        match = select_pattern.match(query)
        if not match:
            return False

        return True

    @staticmethod
    def create(connection, marked_query, values=[], autocommit=True):
        if PrepareStatementBuilder.is_select(marked_query):
            return SelectPreparedStatement(connection,
                                           marked_query, values)

        return PreparedStatement(connection, marked_query,
                                 values, autocommit)


class PrepareStatementBuilderDryRun(PrepareStatementBuilder):

    @staticmethod
    def create(connection, marked_query, values=[], autocommit=True):
        if PrepareStatementBuilder.is_select(marked_query):
            return SelectPreparedStatement(connection, marked_query, values)

        return DryRunPreparedStatement(connection, marked_query, values)


class AutoCommitBehaviour(object):

    class AutoComit(object):
        def __init__(self, connection):
            self._connection = connection

        def try_commit(self):
            self._connection.commit()

    class FakeComit(object):
        def __init__(self, connection):
            pass

        def try_commit(self):
            return

    @staticmethod
    def create(connection, autocommit):
        if autocommit:
            return AutoCommitBehaviour.AutoComit(connection)
        return AutoCommitBehaviour.FakeComit(connection)


class MarkedQuery(object):

    DBAPI2_MARK = '%s'
    PYODBC_MARK = '?'

    @staticmethod
    def build_sql(query, values):
        sql = query[:]
        for value in values:
            sql = sql.replace(MarkedQuery.PYODBC_MARK, str(value))

        return sql

    @staticmethod
    def check_values(query, values):
        n_values = len(values)
        n_marks  = query.count(MarkedQuery.PYODBC_MARK)

        if n_values >= n_marks:
            return

        raise MissingValues(
            'Missing [%s] values for complete the query [%s]' % \
                ((n_marks - n_values), query))

    @staticmethod
    def generate(query):
        return query.replace(MarkedQuery.DBAPI2_MARK,
                             MarkedQuery.PYODBC_MARK)


class PreparedStatement(object):

    SQL_LOG = LogFactory().get_logger(LogFactory.SQL_PREFIX)

    def __init__(self, connection, marked_query, values=(),
                 autocommit=True):
        self._log = LogFactory().get_logger(PreparedStatement.__name__)
        self._connection   = connection
        self._values       = values
        self._marked_query = MarkedQuery.generate(marked_query)
        self._autocommit   = AutoCommitBehaviour.create(self._connection,
                                                        autocommit)
        MarkedQuery.check_values(self._marked_query, values)

    def _trace_sql(self, query):
        PreparedStatement.SQL_LOG.info(query)

    def trace_success(self):
        query = MarkedQuery.build_sql(self._marked_query[:], self._values)
        self._log.debug('Executed statement [%s]' % query)

    def set_values(self, values):
        self._values = values

    def commit(self):
        self._connection.commit()

    def rollback(self):
        self._connection.rollback()

    def execute(self):
        query = MarkedQuery.build_sql(self._marked_query[:], self._values)

        self._trace_sql(query)

        result = self._connection.execute(query)

        ## It is not possible recover the 'prepared' SQL code using:
        ##   result = self._connection.execute(query, self._values)
        ## For this reason the build_sql method is used. That is, the
        ## code generated (and traced) is exactly the code introduced
        ## in the database

        self._autocommit.try_commit()

        return result

    def get_row(self):
        return self._connection.get_row()

    def get_rows(self):
        return self._connection.get_rows()


class DryRunPreparedStatement(PreparedStatement):
    def __init__(self, connection, marked_query, values=[]):
        PreparedStatement.__init__(self, connection, marked_query, values)

    def execute(self):
        query = MarkedQuery.build_sql(self._marked_query[:], self._values)
        self._trace_sql(query)
        self.trace_success()

        return True


class SelectPreparedStatement(PreparedStatement):
    def __init__(self, connection, marked_query, values=[]):
        PreparedStatement.__init__(self, connection, marked_query, values)

    def trace_success(self):
        query = MarkedQuery.build_sql(self._marked_query[:], self._values)
        self._log.trace('Recovered [%d] rows using: [%s]' % \
                            (self._cursor.rowcount, query))
